#!/usr/bin/env python3

from . import plt, layer, optim, act, utils
from .LearningCurveLogger import LearningCurveLogger

from .ClassificationReporter import ClassificationReporter

from .ClassifierServer import ClassifierServer
# from ._switchers import switch_layer, switch_act, switch_optim
from .loss.MultiTaskLoss import MultiTaskLoss
from .EarlyStopping import EarlyStopping



#!/usr/bin/env python3
# Time-stamp: "2021-12-21 21:41:38 (ywatanabe)"

import torch
import torch.nn as nn
import mngs

from functools import partial


class FeatureExtractor(nn.Module):
    def __init__(
        self,
        samp_rate,
        features_list=[
            "mean",
            "std",
            "skewness",
            "kurtosis",
            "median",
            "q25",
            "q75",
            "rms",
            "rfft_bands",
            "beyond_r_sigma_ratio",
        ],
        batch_size=None,
    ):
        super().__init__()

        self.func_dict = dict(
            mean=mngs.dsp.mean,
            std=mngs.dsp.std,
            skewness=mngs.dsp.skewness,
            kurtosis=mngs.dsp.kurtosis,
            median=mngs.dsp.median,
            q25=mngs.dsp.q25,
            q75=mngs.dsp.q75,
            rms=mngs.dsp.rms,
            rfft_bands=partial(mngs.dsp.rfft_bands, samp_rate=samp_rate),
            beyond_r_sigma_ratio=mngs.dsp.beyond_r_sigma_ratio,
        )

        self.features_list = features_list

        self.batch_size = batch_size

    def forward(self, x):
        if self.batch_size is None:
            conc = torch.cat(
                [self.func_dict[f_str](x) for f_str in self.features_list], dim=-1
            )
            return conc
        else:
            conc = []
            n_batches = len(x) // self.batch_size + 1
            for i_batch in range(n_batches):
                try:
                    start = i_batch * self.batch_size
                    end = (i_batch + 1) * self.batch_size
                    conc.append(
                        torch.cat(
                            [
                                self.func_dict[f_str](x[start:end].cuda())
                                for f_str in self.features_list
                            ],
                            dim=-1,
                        ).cpu()
                    )
                except Exception as e:
                    print(e)
            return torch.cat(conc)


def main():
    BS = 32
    N_CHS = 19
    SEQ_LEN = 2000
    SAMP_RATE = 1000
    NYQ = int(SAMP_RATE / 2)

    x = torch.randn(BS, N_CHS, SEQ_LEN).cuda()

    m = FeatureExtractor(SAMP_RATE, batch_size=8)

    out = m(x)  # 15 features


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from .load import get_data_path_from_a_package, load, load_yaml_as_an_optuna_dict, load_study_rdb
from .save import is_listed_X, save, save_listed_dfs_as_csv, save_listed_scalars_as_csv, save_optuna_study_as_csv_and_pngs
from .path import (
    get_this_fpath,
    mk_spath,
    find_the_git_root_dir,
    split_fpath,
    touch,
)



#!/usr/bin/env python3
# Time-stamp: "2021-11-30 17:09:22 (ylab)"

import matplotlib.pyplot as plt
import numpy as np
import pyro
import torch
import torch.nn.functional as F
import torchaudio
from pyro.ops.stats import autocorrelation as pyro_acf
from scipy.signal import chirp, firwin
from functools import partial
from mngs.dsp.HilbertTransformation import HilbertTransformer


## Global definitions
BANDS_LIM_HZ_DICT = {
    "delta": [0.5, 4],
    "theta": [4, 8],
    "lalpha": [8, 10],
    "halpha": [10, 13],
    "beta": [13, 32],
    "gamma": [32, 75],
}


def demo_sig(batch_size=64, n_chs=19, samp_rate=1000, len_sec=10, freqs_hz=[2, 5, 10]):
    time = torch.arange(0, len_sec, 1 / samp_rate)
    sig = torch.vstack([torch.sin(f * 2 * torch.pi * time) for f in freqs_hz]).sum(
        dim=0
    )
    sig = sig.unsqueeze(0).unsqueeze(0).repeat(batch_size, n_chs, 1)
    return sig


def phase(x, axis=-1):
    analytical_x = hilbert(x, axis=-1)
    return analytical_x[..., 0]


def phase_band(x, samp_rate, band_str="delta", l=None, h=None):
    """
    Example:
        x = mngs.dsp.demo_sig(x)
        mngs.dsp.phase_band(x, samp_rate=1000, band_str="delta")

        x = mngs.dsp.demo_sig(x)
        mngs.dsp.phase_band(x, samp_rate=1000, band_str=None, l=0.5, h=4)

    Bands definitions:
        BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }
    """
    is_band_passed = (band_str is not None) & (l is None) & (h is None)
    is_lh_passed = (band_str is None) & (l is not None) & (h is not None)
    assert is_band_passed or is_lh_passed

    if is_band_passed:
        l, h = BANDS_LIM_HZ_DICT[band_str]

    x = bandpass(x, samp_rate, low_hz=l, high_hz=h)
    return phase(x)


def amp(x, axs=-1):
    analytical_x = hilbert(x, axis=-1)
    return analytical_x[..., 1]


def amp_band(x, samp_rate, band_str="delta", l=None, h=None):
    """
    Example:
        x = mngs.dsp.demo_sig(x)
        mngs.dsp.amp_band(x, samp_rate=1000, band_str="delta")

        x = mngs.dsp.demo_sig(x)
        mngs.dsp.amp_band(x, samp_rate=1000, band_str=None, l=0.5, h=4)

    Bands definitions:
        BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }
    """
    is_band_passed = (band_str is not None) & (l is None) & (h is None)
    is_lh_passed = (band_str is None) & (l is not None) & (h is not None)
    assert is_band_passed or is_lh_passed

    if is_band_passed:
        l, h = BANDS_LIM_HZ_DICT[band_str]

    x = bandpass(x, samp_rate, low_hz=l, high_hz=h)
    return amp(x)


def hilbert(x, axis=-1):
    if axis == -1:
        axis = x.ndim - 1
    return HilbertTransformer(axis=axis).to(x.device)(x)


def fft(x, samp_rate):
    fn = partial(_fft_1d, samp_rate=samp_rate, return_freq=False)
    fft_coef = _apply_to_the_time_dim(fn, x)
    freq = torch.fft.fftfreq(x.shape[-1], d=1 / samp_rate)
    return fft_coef, freq


def _fft_1d(x, samp_rate, return_freq=True):
    freq = torch.fft.fftfreq(x.shape[-1], d=1 / samp_rate)
    fft_coef = torch.fft.fft(x)  # [:, :, :nyq]

    if return_freq:
        return fft_coef, freq
    else:
        return fft_coef


def rfft_bands(
    x, samp_rate, bands_str=["delta", "theta", "lalpha", "halpha", "beta", "gamma"]
):
    """
    Returns mean absolute rfft coefficients of bands.
    Bands' definitions are as follows.

    BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }
    """
    coef, freq = rfft(x, samp_rate)

    coef_bands_abs_mean = []
    for band_str in bands_str:
        low, high = BANDS_LIM_HZ_DICT[band_str]
        indi_band = (low <= freq) & (freq <= high)
        coef_band_abs_mean = coef[..., indi_band].abs().mean(dim=-1)
        coef_bands_abs_mean.append(coef_band_abs_mean)

    return torch.stack(coef_bands_abs_mean, dim=-1)


def rfft(x, samp_rate):
    fn = partial(_rfft_1d, samp_rate=samp_rate, return_freq=False)
    fft_coef = _apply_to_the_time_dim(fn, x)
    freq = torch.fft.rfftfreq(x.shape[-1], d=1 / samp_rate)
    return fft_coef, freq


def _rfft_1d(x, samp_rate, return_freq=True):
    freq = torch.fft.rfftfreq(x.shape[-1], d=1 / samp_rate)
    fft_coef = torch.fft.rfft(x)  # [:, :, :nyq]

    if return_freq:
        return fft_coef, freq
    else:
        return fft_coef


def bandstop(x, samp_rate, low_hz=55, high_hz=65):
    fn = partial(_bandstop_1d, samp_rate=samp_rate, low_hz=low_hz, high_hz=high_hz)
    return _apply_to_the_time_dim(fn, x)


def _bandstop_1d(x, samp_rate, low_hz=55, high_hz=65):
    fft_coef, freq = _rfft_1d(x, samp_rate)
    indi_to_cut = (low_hz < freq) & (freq < high_hz)
    fft_coef[indi_to_cut] = 0
    return torch.fft.irfft(fft_coef)


def bandpass(x, samp_rate, low_hz=55, high_hz=65):
    fn = partial(_bandpass_1d, samp_rate=samp_rate, low_hz=low_hz, high_hz=high_hz)
    return _apply_to_the_time_dim(fn, x)


def _bandpass_1d(x, samp_rate, low_hz=55, high_hz=65):
    fft_coef, freq = _rfft_1d(x, samp_rate)
    indi_to_cut = (freq < low_hz) + (high_hz < freq)
    fft_coef[indi_to_cut] = 0
    return torch.fft.irfft(fft_coef)


def spectrogram(x, fft_size, device="cuda"):
    """
    Short-time FFT for signals.

    Input: [BS, N_CHS, SEQ_LEN]
    Output: [BS, N_CHS, fft_size//2 + 1, ?]

    spect = spectrogram(x, 50)
    print(spect.shape)
    """

    transform = torchaudio.transforms.Spectrogram(n_fft=fft_size).to(device)
    spectrogram = transform(x)
    return spectrogram


def mean(x):
    return x.mean(-1, keepdims=True)


def std(x):
    return x.std(-1, keepdims=True)


def zscore(x):
    _mean = mean(x)
    diffs = x - _mean
    var = torch.mean(torch.pow(diffs, 2.0), dim=-1, keepdims=True)
    std = torch.pow(var, 0.5)
    return diffs / std


def kurtosis(x):
    zscores = zscore(x)
    kurtoses = torch.mean(torch.pow(zscores, 4.0), dim=-1, keepdims=True) - 3.0
    return kurtoses


def skewness(x):
    zscores = zscore(x)
    return torch.mean(torch.pow(zscores, 3.0), dim=-1, keepdims=True)


def median(x):
    return torch.median(x, dim=-1, keepdims=True)[0]


def q25(x, q=0.25):
    return torch.quantile(x, q, dim=-1, keepdims=True)


def q75(x, q=0.75):
    return torch.quantile(x, q, dim=-1, keepdims=True)


def rms(x):
    return torch.square(x).sqrt().mean(dim=-1, keepdims=True)


def beyond_r_sigma_ratio(x, r=2.0):
    sigma = std(x)
    return (sigma < x).float().mean(dim=-1, keepdims=True)


def acf(x):
    return pyro_acf(x, dim=-1)


def _apply_to_the_time_dim(fn, x):
    """
    x: [BS, N_CHS, SEQ_LEN]
    When fn(x[0,0]) works, _apply_to_the_time_dim(fn, x) works.
    """
    shape = x.shape
    x = x.reshape(-1, shape[-1])
    dim = 0
    applied = torch.stack([fn(x_i) for x_i in torch.unbind(x, dim=dim)], dim=dim)
    return applied.reshape(shape[0], shape[1], -1)


def _test_notch_filter_1d():
    time = torch.linspace(0, 1, 999)
    sig = (
        torch.cos(60 * 2 * torch.pi * time)
        + torch.cos(200 * 2 * torch.pi * time)
        + torch.cos(300 * 2 * torch.pi * time)
    )

    sig_filted = notch_filter_1d(sig, SAMP_RATE, cutoff_hz=60, width_hz=5)
    fig, axes = plt.subplots(4, 1)
    axes[0].plot(sig, label="sig")
    fft_coef_sig, freq_sig = _rfft_1d(sig, SAMP_RATE)
    axes[1].plot(freq_sig, fft_coef_sig.abs(), label="fft_coef_sig")
    axes[2].plot(sig_filted, label="sig_filted")

    fft_coef_sig_filted, freq_sig_filted = _rfft_1d(sig_filted, SAMP_RATE)
    axes[3].plot(
        freq_sig_filted, fft_coef_sig_filted.abs(), label="fft_coef_sig_filted"
    )
    for ax in axes:
        ax.legend()
    fig.show()


def test_phase_amp_bandpass():
    samp_rate = 1000
    len_seq = 10
    time = torch.arange(0, len_seq, 1 / samp_rate)
    freqs_hz = [2, 5, 10]
    sig = torch.vstack([torch.sin(f * 2 * torch.pi * time) for f in freqs_hz]).sum(
        dim=0
    )
    # sig = _bandstop_1d(sig, samp_rate, low_hz=4, high_hz=12)
    sig = sig.unsqueeze(0).unsqueeze(0).repeat(BS, N_CHS, 1)
    sig = bandpass(sig, samp_rate, low_hz=0, high_hz=3)

    phase = _phase_1d(sig)
    amp = _amp_1d(sig)

    fig, axes = plt.subplots(3, 1, sharex=True, sharey=True)
    axes[0].plot(sig[0, 0], label="sig")
    axes[0].legend()
    axes[1].plot(phase[0, 0], label="phase")
    axes[1].legend()
    axes[2].plot(amp[0, 0], label="amp")
    axes[2].legend()
    fig.show()


if __name__ == "__main__":
    SAMP_RATE = 1000
    x = demo_sig(freqs_hz=[2, 3, 5, 10], samp_rate=SAMP_RATE, len_sec=2)
    # x = torch.tensor(chirp(time, 3, 500, 100))



#!/usr/bin/env python3

# from .path import find_the_git_root_dir, get_this_fpath, mk_spath, split_fpath
# from .load import get_data_path_from_a_package, load
# from .save import is_listed_X, save, save_listed_dfs_as_csv, save_listed_scalars_as_csv
from ..io.__init__ import *
from .cuda_collect_env import main as cuda_collect_env
from .debug import *
from .latex import add_hat_in_the_latex_style, to_the_latex_style
from .mat2py import *
from .misc import (connect_nums, connect_strs, copy_files, copy_the_file,
                   decapitalize, fmt_size, grep, is_defined_global,
                   is_defined_local, is_later_or_equal, isclose, listed_dict,
                   merge_dicts_wo_overlaps, pop_keys, search, squeeze_spaces,
                   take_the_closest, is_nan)
from .pandas import col_to_last, col_to_top, force_dataframe, merge_columns
from .repro import *
from .TimeStamper import *
from .torch import *



#!/usr/bin/env python

import numpy as np
import torch  # 1.7.1
import torch.nn as nn
from torch.fft import fft, ifft


class BaseHilbertTransformer(nn.Module):
    # https://github.com/Project-MONAI/MONAI/blob/master/monai/networks/layers/simplelayers.py
    def __init__(self, axis=2, n=None):
        super().__init__()
        self.axis = axis
        self.n = n

    def transform(self, x):
        """
        Args:
            x: Tensor or array-like to transform.
               Must be real and in shape ``[Batch, chns, spatial1, spatial2, ...]``.
        Returns:
            torch.Tensor: Analytical signal of ``x``,
                          transformed along axis specified in ``self.axis`` using
                          FFT of size ``self.N``.
                          The absolute value of ``x_ht`` relates to the envelope of ``x``
                          along axis ``self.axis``.
        """

        n = x.shape[self.axis] if self.n is None else self.n

        # Create frequency axis
        f = torch.cat(
            [
                torch.true_divide(
                    torch.arange(0, (n - 1) // 2 + 1, device=x.device), float(n)
                ),
                torch.true_divide(
                    torch.arange(-(n // 2), 0, device=x.device), float(n)
                ),
            ]
        )

        xf = fft(x, n=n, dim=self.axis)

        # Create step functionb
        u = torch.heaviside(f, torch.tensor([0.5], device=f.device))
        u = torch.as_tensor(u, dtype=x.dtype, device=u.device)
        new_dims_before = self.axis
        new_dims_after = len(xf.shape) - self.axis - 1
        for _ in range(new_dims_before):
            u.unsqueeze_(0)
        for _ in range(new_dims_after):
            u.unsqueeze_(-1)

        transformed = ifft(xf * 2 * u, dim=self.axis)

        return transformed


class HilbertTransformer(BaseHilbertTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hilbert_transform = self.transform
        self.register_buffer("pi", torch.tensor(np.pi))

    def forward(self, sig):
        sig_comp = self.hilbert_transform(sig)

        pha = self._calc_Arg(sig_comp)
        amp = sig_comp.abs()

        out = torch.cat(
            [
                pha.unsqueeze(-1),
                amp.unsqueeze(-1),
            ],
            dim=-1,
        )

        return out

    def _calc_Arg(self, comp):
        """Calculates argument of complex numbers in (-pi, pi] space.
        Although torch.angle() does not have been implemented the derivative function,
        this function seems to work.
        """
        x, y = comp.real, comp.imag
        Arg = torch.zeros_like(x).type_as(x)

        c1 = x > 0  # condition #1
        Arg[c1] += torch.atan(y / x)[c1]

        c2 = (x < 0) * (y >= 0)
        Arg[c2] += torch.atan(y / x)[c2] + self.pi

        c3 = (x < 0) * (y < 0)
        Arg[c3] += torch.atan(y / x)[c3] - self.pi

        c4 = (x == 0) * (y > 0)
        Arg[c4] += self.pi / 2.0

        c5 = (x == 0) * (y < 0)
        Arg[c5] += -self.pi / 2.0

        c6 = (x == 0) * (y == 0)
        Arg[c6] += 0.0

        return Arg


def _unwrap(x):
    pi = torch.tensor(np.pi)
    y = x % (2 * pi)
    return torch.where(y > pi, 2 * pi - y, y)


def test_hilbert(sig, axis=1, test_layer=True):

    if test_layer:
        ## Hilbert Layer
        hilbert_layer = HilbertTransformer(axis=axis).cuda()
        sig_comp = hilbert_layer(sig)

    else:
        ## Hilbert Transformation
        hilbert = HilbertTransformer(axis=1)
        sig_comp = hilbert.transform(sig)

    print(sig.shape)
    print(sig_comp.shape)

    print(sig.dtype)
    print(sig_comp.dtype)

    print(sig_comp)

    # ## Extract Amplitude and Phase signals
    # amp = sig_comp.abs()
    # pha = _unwrap(sig_comp.angle())

    # instantaneous_freq = (np.diff(pha.cpu()) / (2.0*np.pi) * fs)

    # ## Plot
    # fig, ax = plt.subplots(3, 1)
    # ax[0].set_title('GPU Hilbert Transformation')

    # ax[0].plot(t, sig[0].cpu(), label='original signal')
    # ax[0].plot(t, amp[0].cpu(), label='envelope')
    # ax[0].set_xlabel("time in seconds")
    # ax[0].legend()

    # ax[1].plot(t, pha[0].cpu(), label='phase')
    # ax[1].set_xlabel("time in seconds")
    # ax[1].legend()

    # ax[2].plot(t[1:], instantaneous_freq[0], label='instantenous frequency')
    # ax[2].set_xlabel("time in seconds")
    # ax[2].set_ylim(0.0, 120.0)
    # ax[2].legend()

    # fig.show()

    pha, amp = sig_comp[..., 0], sig_comp[..., 1]
    return pha, amp


def mk_sig():
    from scipy.signal import chirp

    def _mk_sig():
        sig = chirp(t, 20.0, t[-1], 100.0)
        sig *= 1.0 + 0.5 * np.sin(2.0 * np.pi * 3.0 * t)
        return sig

    sig = np.array([_mk_sig() for _ in range(batch_size)]).astype(np.float32)
    sig = torch.tensor(sig).cuda()
    return sig


if __name__ == "__main__":
    # import numpy as np
    import matplotlib
    import matplotlib.pyplot as plt
    from scipy.signal import chirp

    ## Parameters
    duration = 1.0
    fs = 400.0
    samples = int(fs * duration)
    t = np.arange(samples) / fs
    batch_size = 64

    ## Create demo signal
    sig = mk_sig()

    ## Test Code
    pha, amp = test_hilbert(sig, axis=1, test_layer=True)
    # ## EOF



#!/usr/bin/env python3

import collections
import math
import os
import re
import shutil
import time
from bisect import bisect_left, bisect_right
from collections import defaultdict

import mngs
import numpy as np
import pandas as pd
import torch
from natsort import natsorted


################################################################################
## strings
################################################################################
def decapitalize(s):
    if not s:  # check that s is not empty string
        return s
    return s[0].lower() + s[1:]


def connect_strs(strs, filler="_"):  # connect_nums also works as connect_strs
    """
    Example:
        print(connect_strs(['a', 'b', 'c'], filler='_'))
        # 'a_b_c'
    """
    if isinstance(strs, list) or isinstance(strs, tuple):
        connected = ""
        for s in strs:
            connected += filler + s
        return connected[len(filler) :]
    else:
        return strs


def connect_nums(nums, filler="_"):
    """
    Example:
        print(connect_nums([1, 2, 3], filler='_'))
        # '1_2_3'
    """
    if isinstance(nums, list) or isinstance(nums, tuple):
        connected = ""
        for n in nums:
            connected += filler + str(n)
        return connected[len(filler) :]
    else:
        return nums


def squeeze_spaces(string, pattern=" +", repl=" "):
    """Return the string obtained by replacing the leftmost
    non-overlapping occurrences of the pattern in string by the
    replacement repl.  repl can be either a string or a callable;
    if a string, backslash escapes in it are processed.  If it is
    a callable, it's passed the Match object and must return
    a replacement string to be used.
    """
    # return re.sub(" +", " ", string)
    return re.sub(pattern, repl, string)


def search(patterns, strings, only_perfect_match=False):
    """
    regular expression is acceptable for patterns.

    Example:
        patterns = ['orange', 'banana']
        strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        print(search(patterns, strings))
        # ([1, 4, 5], ['orange', 'banana', 'orange_juice'])

        patterns = 'orange'
        strings = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        print(search(patterns, strings))
        # ([1, 5], ['orange', 'orange_juice'])
    """

    ## For single string objects
    def to_list(s_or_p):
        if isinstance(s_or_p, collections.abc.KeysView):
            s_or_p = list(s_or_p)

        elif not isinstance(s_or_p, (list, tuple, pd.core.indexes.base.Index)):
            s_or_p = [s_or_p]

        return s_or_p

    patterns = to_list(patterns)
    strings = to_list(strings)

    ## Main
    if not only_perfect_match:
        indi_matched = []
        for pattern in patterns:
            for i_str, string in enumerate(strings):
                m = re.search(pattern, string)
                if m is not None:
                    indi_matched.append(i_str)
    else:
        indi_matched = []
        for pattern in patterns:
            for i_str, string in enumerate(strings):
                if pattern == string:
                    indi_matched.append(i_str)

    ## Sorts the indices according to the original strings
    indi_matched = natsorted(indi_matched)
    keys_matched = list(np.array(strings)[indi_matched])
    return indi_matched, keys_matched


def grep(str_list, search_key):
    """
    Example:
        str_list = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        search_key = 'orange'
        print(grep(str_list, search_key))
        # ([1, 5], ['orange', 'orange_juice'])
    """
    matched_keys = []
    indi = []
    for ii, string in enumerate(str_list):
        m = re.search(search_key, string)
        if m is not None:
            matched_keys.append(string)
            indi.append(ii)
    return indi, matched_keys


def pop_keys(keys_list, keys_to_pop):
    """
    keys_list = ['a', 'b', 'c', 'd', 'e', 'bde']
    keys_to_pop = ['b', 'd']
    pop_keys(keys_list, keys_to_pop)
    """
    indi_to_remain = [k not in keys_to_pop for k in keys_list]
    keys_remainded_list = list(np.array(keys_list)[list(indi_to_remain)])
    return keys_remainded_list


def fmt_size(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)


################################################################################
## list
################################################################################
def take_the_closest(list_obj, num_insert):
    """
    Assumes list_obj is sorted. Returns the closest value to num.
    If the same number is included in list_obj, the smaller number is returned.

    Example:
        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1.2
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 2

        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 1
    """
    if math.isnan(num_insert):
        closest_num = np.nan
        closest_pos = np.nan

    pos_num_insert = bisect_left(list_obj, num_insert)

    if pos_num_insert == 0:  # When the insertion is at the first position
        closest_num = list_obj[0]
        closest_pos = pos_num_insert  # 0

    if pos_num_insert == len(list_obj):  # When the insertion is at the last position
        closest_num = list_obj[-1]
        closest_pos = pos_num_insert  # len(list_obj)

    else:  # When the insertion is anywhere between the first and the last positions
        pos_before = pos_num_insert - 1

        before_num = list_obj[pos_before]
        after_num = list_obj[pos_num_insert]

        delta_after = abs(after_num - num_insert)
        delta_before = abs(before_num - num_insert)

        if np.abs(delta_after) < np.abs(delta_before):
            closest_num = after_num
            closest_pos = pos_num_insert

        else:
            closest_num = before_num
            closest_pos = pos_before

    return closest_num, closest_pos


################################################################################
## mutable
################################################################################
def isclose(mutable_a, mutable_b):
    return [math.isclose(a, b) for a, b in zip(mutable_a, mutable_b)]


################################################################################
## dictionary
################################################################################
def merge_dicts_wo_overlaps(*dicts):
    merged_dict = {} # init
    for dict in dicts:
        assert mngs.general.search(merged_dict.keys(), dict.keys(), only_perfect_match=True) == ([], [])
        merged_dict.update(dict)
    return merged_dict


def listed_dict(keys=None):  # Is there a better name?
    """
    Example 1:
        import random
        random.seed(42)
        d = listed_dict()
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 1, 0, 4, 3, 3, 2, 1, 10, 8]})

    Example 2:
        import random
        random.seed(42)
        keys = ['a', 'b', 'c']
        d = listed_dict(keys)
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
            d['b'].append(random.randint(0, 10))
            d['c'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 4, 2, 8, 6, 1, 8, 8, 8, 7],
        #                              'b': [1, 3, 1, 1, 0, 3, 9, 3, 6, 9],
        #                              'c': [0, 3, 10, 9, 0, 3, 0, 10, 3, 4]})
    """
    dict_list = defaultdict(list)
    # initialize with keys if possible
    if keys is not None:
        for k in keys:
            dict_list[k] = []
    return dict_list


################################################################################
## variables
################################################################################
def is_defined_global(x_str):
    """
    Example:
        print(is_defined('a'))
        # False

        a = 5
        print(is_defined('a'))
        # True
    """
    return x_str in globals()


def is_defined_local(x_str):
    """
    Example:
        print(is_defined('a'))
        # False

        a = 5
        print(is_defined('a'))
        # True
    """
    return x_str in locals()


# def does_exist(suspicious_var_str):
#     return suspicious_var_str in globals()


################################################################################
## versioning
################################################################################
def is_later_or_equal(package, tgt_version, format="MAJOR.MINOR.PATCH"):
    import mngs
    import numpy as np

    indi, matched = mngs.general.search(["MAJOR", "MINOR", "PATCH"], format.split("."))
    imp_major, imp_minor, imp_patch = [
        int(v) for v in np.array(package.__version__.split("."))[indi]
    ]
    tgt_major, tgt_minor, tgt_patch = [
        int(v) for v in np.array(tgt_version.split("."))[indi]
    ]

    print(
        f"\npackage: {package.__name__}\n"
        f"target_version: {tgt_version}\n"
        f"imported_version: {imp_major}.{imp_minor}.{imp_patch}\n"
    )

    ## Mjorr
    if imp_major > tgt_major:
        return True

    if imp_major < tgt_major:
        return False

    if imp_major == tgt_major:

        ## Minor
        if imp_minor > tgt_minor:
            return True

        if imp_minor < tgt_minor:
            return False

        if imp_minor == tgt_minor:

            ## Patch
            if imp_patch > tgt_patch:
                return True
            if imp_patch < tgt_patch:
                return False
            if imp_patch == tgt_patch:
                return True


################################################################################
## File
################################################################################
def _copy_a_file(src, dst, allow_overwrite=False):
    if src == "/dev/null":
        print(f"\n/dev/null was not copied.\n")

    else:

        if dst.endswith("/"):
            _, src_fname, src_ext = mngs.general.split_fpath(src)
            # src_fname = src + src_ext
            dst = dst + src_fname + src_ext

        if not os.path.exists(dst):
            shutil.copyfile(src, dst)
            print(f'\nCopied "{src}" to "{dst}".\n')

        else:
            if allow_overwrite:
                shutil.copyfile(src, dst)
                print(f'\nCopied "{src}" to "{dst}" (overwritten).\n')

            if not allow_overwrite:
                print(f'\n"{dst}" exists and copying from "{src}" was aborted.\n')


def copy_files(src_files, dists, allow_overwrite=False):
    if isinstance(src_files, str):
        src_files = [src_files]

    if isinstance(dists, str):
        dists = [dists]

    for sf in src_files:
        for dst in dists:
            _copy_a_file(sf, dst, allow_overwrite=allow_overwrite)


def copy_the_file(sdir):  # dst
    __file__ = inspect.stack()[1].filename
    _, fname, ext = mngs.general.split_fpath(__file__)

    dst = sdir + fname + ext

    if "ipython" not in __file__:  # for ipython
        # shutil.copyfile(__file__, dst)
        # print(f"Saved to: {dst}")
        _copy_a_file(__file__, dst)

def is_nan(X):
    if isinstance(X, pd.DataFrame):
        if X.isna().any().any():
            raise ValueError("NaN was found in X")
    elif isinstance(X, np.ndarray):
        if np.isnan(X).any():        
            raise ValueError("NaN was found in X")
    elif torch.is_tensor(X):
        if X.isnan().any():        
            raise ValueError("NaN was found in X")
    elif isinstance(X, (float, int)):
        if math.isnan(X):
            raise ValueError("X was NaN")




#!/usr/bin/env python
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sys
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.autograd import Variable
import time
import math
import matplotlib.pyplot as plt
import pandas as pd
import scipy.signal


class BandPassFilter(nn.Module):
    # https://raw.githubusercontent.com/mravanelli/SincNet/master/dnn_models.py
    def __init__(self, order=None, low_hz=30, high_hz=60, fs=250, n_chs=19):
        super().__init__()
        self.fs = fs
        nyq = fs / 2
        self.order = fs if order is None else order
        self.numtaps = self.order + 1
        filter_npy = scipy.signal.firwin(
            self.numtaps, [low_hz, high_hz], pass_zero="bandpass", fs=fs,
        )

        self.register_buffer('filters',
                             torch.tensor(filter_npy).unsqueeze(0).unsqueeze(0)
        )
        

    def forward(self, x):
        dim = x.ndim
        sig_len_orig = x.shape[-1]

        if dim == 3:
            bs, n_chs, sig_len = x.shape
            x = x.reshape(bs*n_chs, 1, sig_len)
            
        filted = F.conv1d(x, self.filters.type_as(x), padding=int(self.numtaps/2))
        filted = filted.flip(dims=[-1]) # to backward
        filted = F.conv1d(filted, self.filters.type_as(x), padding=int(self.numtaps/2))
        filted = filted.flip(dims=[-1]) # reverse to the original order        

        filted = filted[..., 1:-1]
        # print(self.order, filted.shape[-1])        

        if dim == 3:
            filted = filted.reshape(bs, n_chs, -1)
        
        return filted


class BandPasser_CPU:
    def __init__(self, low_hz=100, high_hz=250, fs=1000):
        from scipy.signal import butter, sosfilt

        self.butter = butter
        self.sosfilt = sosfilt

        self.sos = self.mk_sos(low_hz, high_hz, fs)

    def __call__(self, raw_sig):
        filted = self.sosfilt(self.sos, raw_sig)
        return filted

    def mk_sos(self, low_hz, high_hz, fs, order=5):
        nyq = fs / 2.0
        sos = self.butter(
            order,
            [low_hz / nyq, high_hz / nyq],
            analog=False,
            btype="band",
            output="sos",
        )
        return sos

    
def wavelet(
    wave, samp_rate, f_min=100, f_max=None, plot=False, title=None
):  # for visual checking
    from obspy.signal.tf_misfit import cwt
    from obspy.imaging.cm import obspy_sequential

    dt = 1.0 / samp_rate
    npts = len(wave)
    t = np.linspace(0, dt * npts, npts)
    if f_min == None:
        f_min = 0.1
    if f_max == None:
        f_max = int(samp_rate / 2)
    scalogram = cwt(wave, dt, 8, f_min, f_max)

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        x, y = np.meshgrid(
            t, np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0])
        )

        ax.pcolormesh(x, y, np.abs(scalogram), cmap=obspy_sequential)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Frequency [Hz]")
        ax.set_yscale("log")
        ax.set_title(title)
        ax.set_ylim(f_min, f_max)
        fig.show()

    Hz = pd.Series(np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]))
    df = pd.DataFrame(np.abs(scalogram))
    df["Hz"] = Hz
    df.set_index("Hz", inplace=True)

    return df

    
## from here, not important
def summarize_sinusoidal_waves(T_sec=50.0, fs=250, freqs_hz=[30, 60, 100, 200, 1000]):
    n = int(T_sec * fs)
    t = np.linspace(0, T_sec, n, endpoint=False)
    summed = np.array(
        [
            np.random.rand() * np.sin( (f_hz * t + np.random.rand() ) * (2 * np.pi) )            
            for f_hz in freqs_hz
        ]
    ).sum(axis=0)
    return summed

def prepair_demo_data(T_sec=50.0, fs=250, freqs_hz=[30, 60, 100, 200, 1000], n_chs=19):
    data = np.array([
        summarize_sinusoidal_waves(T_sec=T_sec, fs=fs, freqs_hz=freqs_hz)
        for _ in range(n_chs)
        ])
    return data


class TimeKeeper():
    def __init__(self,):
        self.start_time = time.time()
        
    def __call__(self, message=None):
        self.elapsed = time.time() - self.start_time
        print(message)        
        print('elapsed time: {:.5f} [sec]'.format(self.elapsed))

    def start(self, ):
        self.start_time = time.time()
    

if __name__ == "__main__":
    ## Parameters
    PASS_LOW_HZ = 40
    PASS_HIGH_HZ = 80
    T_sec = 50.0
    fs = 250
    freqs_hz = [30, 60, 100, 200, 1000]
    n_chs = 19
    bs = 64
    i_batch_show = 0
    i_ch_show = 0    
    sig_orig = np.array([
        prepair_demo_data(T_sec=T_sec, fs=fs, freqs_hz=freqs_hz, n_chs=n_chs)
        for _ in range(bs)
    ])
    print('sig_orig_shape: {}'.format(sig_orig.shape))
    tk = TimeKeeper()    
        
    
    ## Original
    title_orig = "Original, #{}, ch{} (Demo Freqs: {} Hz)"\
        .format(i_batch_show, i_ch_show, freqs_hz)
    wv_out_orig = wavelet(sig_orig[i_batch_show, i_ch_show], fs,
                          f_min=1, plot=True, title=title_orig)

    
    ## CPU Bandpass Filtering
    bp_cpu = BandPasser_CPU(low_hz=PASS_LOW_HZ, high_hz=PASS_HIGH_HZ, fs=fs)
    filted_cpu = sig_orig.copy()
    ### calculation start ###
    tk.start()
    for i_batch in range(bs):
        for i_ch in range(n_chs):
            filted_cpu[i_batch, i_ch] = bp_cpu(filted_cpu[i_batch, i_ch])
    tk(message='CPU')
    ### calculation end ###    
    title_filt_cpu = (
        "[CPU] Bandpass Filted, #{}, ch{} (Freqs: (Low_lim, High_lim) = ({}, {}) Hz) (time: {:.5f} [sec])"\
        .format(i_batch_show, i_ch_show, PASS_LOW_HZ, PASS_HIGH_HZ, tk.elapsed
        )
    )
    _wv_out_filted_cpu = wavelet(filted_cpu[i_batch_show, i_ch_show], fs,
                                 f_min=1, plot=True, title=title_filt_cpu)

    
    ## GPU Bandpass Filtering
    BandPassFilter_GPU = BandPassFilter
    bp_gpu = BandPassFilter_GPU(
        low_hz=PASS_LOW_HZ, high_hz=PASS_HIGH_HZ, fs=fs
    ).cuda()
    # sig_torch = torch.tensor(sig_orig).unsqueeze(0).unsqueeze(0).cuda()
    sig_torch = torch.tensor(sig_orig).cuda()
    ### calculation start ###
    tk.start()
    filted_gpu = bp_gpu(sig_torch)
    tk(message='GPU')
    ### calculation end ###    
    title_filt_gpu = (
        "[GPU] Bandpass Filted, #{}, ch{} (Freqs: (Low_lim, High_lim) = ({}, {}) Hz) (time: {:.5f} [sec])"\
        .format(i_batch_show, i_ch_show, PASS_LOW_HZ, PASS_HIGH_HZ, tk.elapsed
        )
    )
    _wv_out_filted_gpu = wavelet(filted_gpu[i_batch_show, i_ch_show].cpu(), fs,
                                 f_min=1, plot=True, title=title_filt_gpu)

    
    ## EOF



#!/usr/bin/env python3

from ._bandpassfilter import bandpassfilter
from .fft_powers import calc_fft_powers
from .wavelet import wavelet
from .FeatureExtractor import FeatureExtractor
from .feature_extractions import (
    rfft_bands,
    rfft,
    bandstop,
    spectrogram,
    mean,
    std,
    zscore,
    kurtosis,
    skewness,
    median,
    q25,
    q75,
    rms,
    beyond_r_sigma_ratio,
    acf,
    demo_sig,
    phase,
    phase_band,
    amp,
    amp_band,
    hilbert,
    fft,
    bandpass,
)
from ._BANDS_LIM_HZ_DICT import BANDS_LIM_HZ_DICT
